/**
 * Copyright (C) 2010 OSGi fuer Praktiker <info@osgi-buch.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *         http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package biz.gossipmonger.gatekeeper;

import java.io.File;
import java.util.Dictionary;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Map;

import org.osgi.framework.BundleActivator;
import org.osgi.framework.BundleContext;
import org.osgi.service.event.Event;
import org.osgi.service.event.EventAdmin;
import org.osgi.service.event.EventConstants;
import org.osgi.service.event.EventHandler;
import org.osgi.util.tracker.ServiceTracker;

import biz.gossipmonger.enunciator.EnunciatorService;
import biz.gossipmonger.worker.WorkerService;

public class GateKeeper implements BundleActivator, EventHandler {
  static String ENUNCIATE_PREFIX = "GateKeeper: ";
  File watchedDirectory = new File("render_input");
  File archiveDirectory = new File("render_archive");
  ServiceTracker enunciatorTracker, workerTracker, eventAdminTracker;

  public void start(BundleContext context) throws Exception {
    enunciatorTracker = new ServiceTracker(context, EnunciatorService.class.getName(), null);
    enunciatorTracker.open();
    workerTracker = new ServiceTracker(context, WorkerService.class.getName(), null);
    workerTracker.open();
    eventAdminTracker = new ServiceTracker(context, EventAdmin.class.getName(), null);
    eventAdminTracker.open();
    String[] topics = new String[] { "biz/gossipmonger/jobtimer" };
    Dictionary<String, Object> dict = new Hashtable<String, Object>();
    dict.put(EventConstants.EVENT_TOPIC, topics);
    context.registerService(EventHandler.class.getName(), this, dict);
  }

  public void stop(BundleContext context) throws Exception {
    if (enunciatorTracker != null) {
      enunciatorTracker.close();
    }
    if (workerTracker != null) {
      workerTracker.close();
    }
    if (eventAdminTracker != null) {
      eventAdminTracker.close();
    }
  }

  public void handleEvent(Event event) {
    prepareDirectory(watchedDirectory);
    prepareDirectory(archiveDirectory);
    String list[] = watchedDirectory.list();
    for (String filename : list) {
      processFile(new File(watchedDirectory, filename));
    }
  }

  private void processFile(File input) {
    enunciate("Processing " + input.getName());
    WorkerService worker = (WorkerService) workerTracker.getService();
    if (worker == null) {
      enunciate("No worker service.");
    } else {
      File output = worker.process(input);
      postResult(input, output);
      if (output != null) {
        enunciate("Moving processed file to archive dir " + archiveDirectory);
        if (input.renameTo(new File(archiveDirectory, input.getName())) == false) {
          enunciate("Error moving file. Please check write permissions.");
        }
      }
    }
  }
  
  private void postResult(File input, File output) {
    final String EVENT_TOPIC = "biz/gossipmonger/gatekeeper";
    final String EVENT_KEY_IN = "input";
    final String EVENT_KEY_OUT = "output";
    EventAdmin ea = (EventAdmin) eventAdminTracker.getService();
    if (ea != null) {
      Map<String,String> info = new HashMap<String,String>();
      info.put(EVENT_KEY_IN, input.getAbsolutePath());
      info.put(EVENT_KEY_OUT, output.getAbsolutePath());
      Event ev = new Event(EVENT_TOPIC, info);
      ea.postEvent(ev);
    } else {
      enunciate("No EventAdmin");
    }
  }

  private void prepareDirectory(File dir) {
    if (!dir.exists()) {
      if (!dir.mkdirs()) {
        throw new RuntimeException("Can't create folder " + dir);
      }
      enunciate("Created directory " + dir.getAbsolutePath());
    }
    if (!dir.isDirectory()) {
      throw new RuntimeException(dir + " is not a directory");
    }
  }

  private void enunciate(String text) {
    if (enunciatorTracker == null) {
      System.out.println(ENUNCIATE_PREFIX+"No EnunciatorService tracker present; Message is: " + text);
    } else {
      EnunciatorService enunciator = null;
      try {
        Object obj = enunciatorTracker.waitForService(500);
        enunciator = (EnunciatorService)obj;
      } catch (InterruptedException e) {
        System.out.println(ENUNCIATE_PREFIX+"was interrupted while waiting for Enunciator");
      }
      if (enunciator != null) {
        enunciator.enunciate(ENUNCIATE_PREFIX+text);
      } else {
        System.out.println(ENUNCIATE_PREFIX+"(no EnunciatorService): " + text);
      }
    }
  }
}
