/**
 * Copyright (C) 2010 OSGi fuer Praktiker <info@osgi-buch.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *         http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package biz.gossipmonger.launcher;

import java.io.BufferedReader;
import java.io.File;
import java.io.IOException;
import java.io.InputStreamReader;
import java.lang.reflect.Field;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.jar.JarFile;

import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;
import org.osgi.framework.BundleException;
import org.osgi.framework.Constants;
import org.osgi.framework.FrameworkEvent;
import org.osgi.framework.launch.Framework;
import org.osgi.framework.launch.FrameworkFactory;

public final class Launcher {
  private static final String APP_BUNDLES = "biz.gossipmonger.launcher.bundles";
  private BundleContext context;

  public static void main(String[] args) {
    try {
      Launcher launcher = new Launcher();
      launcher.launch(args);
    } catch (Exception e) {
      e.printStackTrace();
    }
  }

  private void launch(String[] args) throws Exception {
    // Create and start the framework
    FrameworkFactory factory = getFrameworkFactory();
    Map<String, String> map = getFrameworkSettings(args);
    Framework framework = factory.newFramework(map);
    framework.start();

    // Activate the infrastructure
    context = framework.getBundleContext();
    activate(getInfrastructureBundles(factory.getClass().getName()));

    // Debug property set => activate debug bundles
    String dbg = context.getProperty("biz.gossipmonger.launcher.debug");
    if (dbg != null && dbg.equalsIgnoreCase("true")) {
      activate(getDebugBundles());
    }
    
    // Finally activate the application bundles
    String appsURL = context.getProperty(APP_BUNDLES);
    activate(getApplicationBundles(appsURL));

    // Wait until the framework stopped.
    printFrameworkEvent(framework.waitForStop(0));
  }

  private void printFrameworkEvent(FrameworkEvent fe) {
    String eventType = "Unknown type";
    Field[] fields = fe.getClass().getDeclaredFields();
    for (Field field : fields) {
      try {
        if (field.getInt(fe) == fe.getType()) {
          eventType = field.getName();
        }
      } catch (Exception e) {
      }
    }
    System.out.println(fe.getSource().toString() + ": " + eventType);
  }

  private List<String> getInfrastructureBundles(String factoryClassName) throws IOException {
    File libDir = new File(new File("").getAbsoluteFile(), "lib");
    String path = libDir.getAbsolutePath();
    System.out.println("Looking for infrastructure bundles in " + path);
    if (libDir.exists() == false) {
      throw new IOException(path + " doesn't exist!");
    }
    List<String> locations = new ArrayList<String>();
    String factoryEntry = factoryClassName.replace('.', '/') + ".class";
    for (File file : libDir.listFiles()) {
      if (isInfrastructureBundle(file, factoryEntry)) {
        locations.add("file:" + file.getPath());
      }
    }
    return locations;
  }

  private boolean isInfrastructureBundle(File file, String factoryEntry) {
    try {
      JarFile jarFile = new JarFile(file);
      return jarFile.getEntry(factoryEntry) == null;
    } catch (IOException e) {
      System.err.println(file + " is no JAR.");
    }
    return false;
  }

  private void activate(List<String> bundleLocations) throws BundleException {
    List<Bundle> bundles = new ArrayList<Bundle>();
    for (String location : bundleLocations) {
      try {
        System.out.println("Installing "+location);
        Bundle b = context.installBundle(location);
        bundles.add(b);
      } catch (BundleException e) {
        System.err.println("Cannot load "+location);
        e.printStackTrace();
      }
    }
    for (Bundle bundle : bundles) {
      // Skip Fragments; they cannot be started
      if (bundle.getHeaders().get(Constants.FRAGMENT_HOST) == null) {
        try {
          bundle.start();
        } catch (BundleException e) {
          System.err.println("Cannot start bundle "+
              bundle.getSymbolicName()+"/"+
              bundle.getVersion());
          e.printStackTrace();
        }
      }
    }
  }

  private FrameworkFactory getFrameworkFactory() throws Exception {
    String filename = "META-INF/services/org.osgi.framework.launch.FrameworkFactory";
    List<String> content = getResourceContent(filename);
    if (content.size() == 0) {
      throw new Exception("No FrameworkFactory defined");
    }
    if (content.size() > 1) {
      System.err.println(filename + " contains more than one entry. Will use the first one.");
    }
    return (FrameworkFactory) Class.forName(content.get(0)).newInstance();
  }

  private Map<String, String> getFrameworkSettings(String[] args) throws IOException {
    String path = "META-INF/settings/FrameworkSettings";
    List<String> content = getResourceContent(path);
    if (content.size() == 0) {
      System.out.println("No framework settings found");
    }
    Map<String, String> result = new HashMap<String, String>();
    for (String line : content) {
      String[] parts = line.split("=", 2);
      if (parts.length == 2) {
        result.put(parts[0].trim(), parts[1].trim());
      } else {
        System.err.println("Skipping \"" + line + "\"");
      }
    }
    if (args.length > 0) {
      result.put(APP_BUNDLES, args[0]);
    }
    return result;
  }

  private List<String> getDebugBundles() throws IOException {
    return getResourceContent("META-INF/settings/DebugBundles");
  }

  private List<String> getApplicationBundles(String appsURL) throws IOException {
    if (appsURL == null) {
      return getResourceContent("META-INF/settings/ApplicationBundles");
    }
    try {
      return getUrlContent(new URL(appsURL));
    } catch (MalformedURLException e) {
      return getUrlContent(new File(appsURL).toURI().toURL());
    }
  }

  private List<String> getResourceContent(String filename) throws IOException {
    URL url = getClass().getClassLoader().getResource(filename);
    return getUrlContent(url);
  }

  private List<String> getUrlContent(URL url) throws IOException {
    System.out.println("Reading " + url);
    List<String> result = new ArrayList<String>();
    BufferedReader br = new BufferedReader(new InputStreamReader(url.openStream()));
    try {
      for (String s = br.readLine(); s != null; s = br.readLine()) {
        s = s.trim();
        if ((s.length() > 0) && (s.charAt(0) != '#')) {
          result.add(s);
        }
      }
    } finally {
      br.close();
    }
    return result;
  }
}
